"""
Cloud Formation Template

"""
import json


class Template(object):
    """Used to create a Cloud Formation configuration template"""
    PARENT_CONFIG_PREFIX = ''
    STACK_TYPE = ''

    def __init__(self, name):
        """Create a new Cloud Formation configuration template instance"""
        self._description = 'Formulary created Cloud Formation stack'
        self._name = name

        self._conditions = {}
        self._mappings = {}
        self._metadata = {}
        self._outputs = {}
        self._parameters = {}
        self._resources = {}

    def add_resource(self, resource_id, resource):
        """Add a resource to the template

        :param str resource_id: The camelCase resource_id
        :param Resource resource: The resource to add

        """
        self._resources[resource_id] = resource

    def as_json(self, indent=2):
        """Return the cloud-formation template as JSON

        :param int indent: spaces to indent the JSON by

        """
        resources = dict({})
        for key, value in self._resources.items():
            resources[key] = value.as_dict()
        return json.dumps({'AWSTemplateFormatVersion': '2010-09-09',
                           'Conditions': self._conditions,
                           'Description': self._description,
                           'Mappings': self._mappings,
                           'Metadata': self._metadata,
                           'Outputs': self._outputs,
                           'Parameters': self._parameters,
                           'Resources': resources},
                          indent=indent, sort_keys=True)

    @property
    def name(self):
        """Return the template name

        :rtype: str

        """
        return self._name

    def set_description(self, description):
        """Set the template description

        :param str description: The description value
        """
        self._description = description

    def update_mappings(self, mappings):
        """Update the mappings with values from another dict.

        :param dict mappings: Additional mappings to add

        """
        self._mappings.update(mappings)

    def update_resources(self, resources):
        """Update the resources with the resources from another dict.

        :param dict resources: Additional resources to add

        """
        self._resources.update(resources)



"""
Builders construct the data structures required to build cloud formation
templates.

"""
from formulary.builders.environment import Environment
from formulary.builders.rds import RDS
from formulary.builders.service import Service



"""
Cloud Formation VPC Resources

"""
from formulary import base
from formulary import utils


class DHCPOptions(base.Resource):
    def __init__(self, domain_name, name_servers, ntp_servers):
        super(DHCPOptions, self).__init__('AWS::EC2::DHCPOptions')
        self._properties['DomainName'] = domain_name
        self._properties['DomainNameServers'] = name_servers
        self._properties['NtpServers'] = ntp_servers


class DHCPOptionsAssociation(base.Resource):
    def __init__(self, dhcp_id, vpc_id):
        super(DHCPOptionsAssociation,
              self).__init__('AWS::EC2::VPCDHCPOptionsAssociation')
        self._properties['DhcpOptionsId'] = {'Ref': dhcp_id}
        self._properties['VpcId'] = {'Ref': vpc_id}


class Gateway(base.Resource):
    def __init__(self):
        super(Gateway, self).__init__('AWS::EC2::InternetGateway')


class GatewayAttachment(base.Resource):
    def __init__(self, vpc_id, gateway_id):
        super(GatewayAttachment,
              self).__init__('AWS::EC2::VPCGatewayAttachment')
        self._properties['InternetGatewayId'] = {'Ref': gateway_id}
        self._properties['VpcId'] = {'Ref': vpc_id}


class NetworkACL(base.Resource):
    def __init__(self, vpc_name, vpc_id):
        super(NetworkACL, self).__init__('AWS::EC2::NetworkAcl')
        self._name = '{0}-acl'.format(vpc_name)
        self._properties['VpcId'] = {'Ref': vpc_id}


class NetworkACLEntry(base.Resource):
    def __init__(self, acl_id, cidr_block, rule_number, action, egress, ports):
        super(NetworkACLEntry, self).__init__('AWS::EC2::NetworkAclEntry')
        protocol, from_port, to_port = utils.parse_port_value(ports, -1)
        self._properties['NetworkAclId'] = {'Ref': acl_id}
        self._properties['CidrBlock'] = cidr_block
        self._properties['RuleNumber'] = rule_number
        self._properties['Protocol'] = protocol
        self._properties['RuleAction'] = action
        self._properties['Egress'] = egress
        self._properties['PortRange'] = {'From': from_port, 'To': to_port}


class Route(base.Resource):
    def __init__(self, route_table_id, cidr_block, gateway_id, depends_on):
        super(Route, self).__init__('AWS::EC2::Route')
        self._properties['RouteTableId'] = {'Ref': route_table_id}
        self._properties['DestinationCidrBlock'] = cidr_block
        self._properties['GatewayId'] = {'Ref': gateway_id}
        self._attributes['DependsOn'] = depends_on


class RouteTable(base.Resource):
    def __init__(self, vpc_id):
        super(RouteTable, self).__init__('AWS::EC2::RouteTable')
        self._properties['VpcId'] = {'Ref': vpc_id}


class Subnet(base.Resource):
    def __init__(self, vpc_name, subnet, vpc_id, az, cidr_block):
        super(Subnet, self).__init__('AWS::EC2::Subnet')
        self._name = '{0}{1}-subnet'.format(vpc_name, subnet)
        self._properties['AvailabilityZone'] = az
        self._properties['CidrBlock'] = cidr_block
        self._properties['VpcId'] = {'Ref': vpc_id}


class SubnetRouteTableAssociation(base.Resource):
    def __init__(self, subnet_id, route_table_id):
        super(SubnetRouteTableAssociation,
              self).__init__('AWS::EC2::SubnetRouteTableAssociation')
        self._properties['SubnetId'] = {'Ref': subnet_id}
        self._properties['RouteTableId'] = {'Ref': route_table_id}


class VPC(base.Resource):
    def __init__(self, name, dns_support, dns_hostnames, cidr_block):
        super(VPC, self).__init__('AWS::EC2::VPC')
        self._name = name
        self._properties['EnableDnsSupport'] = dns_support
        self._properties['EnableDnsHostnames'] = dns_hostnames
        self._properties['CidrBlock'] = cidr_block



"""
Base Classes

"""
import logging

from formulary import utils

LOGGER = logging.getLogger(__name__)


class Builder(object):
    def __init__(self, config, name, environment, mappings):
        self._config = self._flatten_config(config, environment)
        self._name = name
        self._environment = environment
        self._mappings = mappings
        self._resources = {}

    @property
    def environment(self):
        """Return the environment value

        :rtype: str

        """
        return self._environment

    @property
    def name(self):
        """Return the builder's resource name

        :rtype: str

        """
        return self._name

    @property
    def resources(self):
        """Return the resource dictionary for the builder"""
        return dict(self._resources)

    def _add_resource(self, name, resource):
        """Add a resource to the template, returning the cloud formation
        template reference name for the resource.

        :param str name: The underscore delimited resource name
        :param Resource resource: The resource to add
        :rtype: str

        """
        resource_id = utils.camel_case(name)
        self._resources[resource_id] = resource
        return resource_id

    @staticmethod
    def _flatten_config(config, environment):
        for key, value in config.items():
            if value == environment:
                config[key] = config[key][environment]
        return config

    def _maybe_replace_with_mapping(self, value):
        """If the value is a ^map macro, replace the with the value from the
        mappings dict. For example, if the value is ``^map Foo.Bar.Baz``
        return self._mappings['Foo']['Bar']['Baz'].

        :param str value: The value to check for the map macro
        :rtype: str|dict
        :raises: ValueError

        """
        if value.startswith('^map '):
            ref = value[5:].split('.')
            if len(ref) != 3:
                raise ValueError('Invalid map reference: {}'.format(value[5:]))
            return self._mappings[ref[0]][ref[1]][ref[2]]
        return value


class Property(object):
    def __init__(self):
        self._values = dict()

    def as_dict(self):
        values = dict(self._values)
        for key, value in self._values.items():
            if value is None or value == [] or value == {}:
                del values[key]
        return values


class Resource(object):
    """Cloud Formation Resource

    Represents a Cloud Formation resource as a class

    """
    def __init__(self, resource_type):
        """Create a new resource specifying the resource type

        :param str resource_type: Resource type such as ``AWS::EC2::Route``

        """
        self._attributes = {}
        self._name = None
        self._properties = {}
        self._tags = {}
        self._type = resource_type

    def add_attribute(self, name, value):
        """Add a top-level attribute to the resource such as ``DependsOn`` in
        ``AWS::EC2::Route``

        :param str name: The attribute name
        :param str value: The attribute value

        """
        self._attributes[name] = value

    def add_property(self, name, value):
        """Add a property to the resource

        :param str name: The name of the property
        :param str|dict|list|bool|int value: The property value

        """
        self._properties[name] = value

    def add_tag(self, name, value):
        """Add a tag to the resource

        :param str name: The name of the tag
        :param str value: The tag value

        """
        self._tags[name] = value

    def as_dict(self):
        """Return the resource as a dictionary for building the cloud-formation
        configuration.

        :return: dict

        """
        self._prune_empty_properties()
        dict_val = dict({'Type': self._type, 'Properties': self._properties})
        if self._tags:
            dict_val['Properties']['Tags'] = []
            for key, value in self._tags.items():
                dict_val['Properties']['Tags'].append({'Key': key,
                                                       'Value': value})
            if self._name:
                dict_val['Properties']['Tags'].append({'Key': 'Name',
                                                       'Value': self._name})
        for key, value in self._attributes.items():
            dict_val[key] = value
        return dict_val

    def set_name(self, name):
        """Set the resource name for use as a tag

        :param str name: The resource name

        """
        self._name = name

    def _prune_empty_properties(self):
        for key, value in list(self._properties.items()):
            if self._properties[key] is None:
                del self._properties[key]



"""
Cloud Formation RDS Resources

"""
from formulary import base


class DBInstance(base.Resource):
    def __init__(self, name, config, subnet, security_grp):
        super(DBInstance, self).__init__('AWS::RDS::DBInstance')
        self._name = name
        self._properties = {
            'AllocatedStorage': config.get('storage-capacity'),
            'AutoMinorVersionUpgrade': config.get('minor-version-upgrade',
                                                  False),
            'BackupRetentionPeriod': config.get('backup-retention'),
            'DBName': config.get('dbname'),
            'DBInstanceClass': config.get('instance-type'),
            'DBInstanceIdentifier': name,
            'DBSubnetGroupName': {'Ref': subnet},
            'Engine': config.get('engine'),
            'EngineVersion': config.get('engine-version'),
            'Iops': config.get('iops'),
            'MasterUsername': config.get('username'),
            'MasterUserPassword': config.get('password'),
            'MultiAZ': config.get('multi-az', True),
            'Port': config.get('port'),
            'PubliclyAccessible': config.get('public', False),
            'VPCSecurityGroups': [security_grp]
        }

        if not config.get('multi-az'):
            self._properties['AvailabilityZone'] =  \
                config.get('availability_zone')

        self.add_attribute('DeletionPolicy',
                           config.get('deletion-policy',
                                      'Delete').capitalize())


class DBSubnetGroup(base.Resource):
    def __init__(self, name, subnets):
        super(DBSubnetGroup, self).__init__('AWS::RDS::DBSubnetGroup')
        self._name = name
        self._properties = {
            'DBSubnetGroupDescription': 'Subnet Group for {0}'.format(name),
            'SubnetIds': subnets
        }



"""
Main Formulary Controller

"""
import logging
from os import path
import sys

import boto.provider
import yaml

from formulary import builders
from formulary import cloudformation
from formulary import stack
from formulary import template

LOGGER = logging.getLogger(__name__)

ACTIONS = {'create', 'update'}
CONFIG_FILES = {'amis.yaml', 'mapping.yaml', 'instances.yaml',
                'environments', 'rds', 'services'}
RESOURCE_TYPES = {'environment', 'service', 'elasticache', 'rds',  'stack'}
STACK_FOLDERS = {'elasticache': 'elasticache',
                 'environment': 'environments',
                 'rds': 'rds',
                 'service': 'services',
                 'stack': 'stacks'}


class Controller(object):
    """The controller implements the top-level application behavior"""

    def __init__(self, config_path, action, environment, resource_type,
                 resource, verbose, dry_run, profile):
        config_path = self._normalize_path(config_path)
        self._validate_arguments(config_path, action, environment,
                                 resource_type, resource)

        self._action = action
        self._config_path = config_path
        self._environment = environment
        self._profile = profile
        self._resource = resource
        self._resource_type = resource_type
        self._verbose = verbose
        self._dry_run = dry_run

        self._amis = self._load_config_file('.', 'amis')
        self._instances = self._load_config_file('.', 'instances')
        self._config = self._load_config()
        self._environment_config = self._load_environment_config()
        self._mappings = self._load_mappings()
        self._template = template.Template(self._template_name)
        self._stack = self._get_stack()

    def execute(self):
        """Create or update a Cloud Formation stack"""
        self._build_template_resources()

        if self._config.get('description'):
            self._template.set_description(self._config['description'])

        template_value = self._template.as_json()
        if self._dry_run:
            print(template_value)
            return

        if self._action == 'create':
            try:
                cloudformation.create_stack(self._region, self._template,
                                            self._profile)
            except boto.provider.ProfileNotFoundError:
                self._error('AWS profile not found')
            print('Stack created')

        elif self._action == 'update':
            try:
                cloudformation.update_stack(self._region, self._template,
                                            self._profile)
            except boto.provider.ProfileNotFoundError:
                self._error('AWS profile not found')
            print('Stack updated')

        result = cloudformation.estimate_stack_cost(self._region,
                                                    self._template,
                                                    self._profile)
        print('Stack cost calculator URL: {0}'.format(result))

    def _build_environment_resources(self):
        builder = builders.Environment(self._config, self._resource,
                                       self._mappings)
        self._template.update_resources(builder.resources)

    def _build_rds_resources(self):
        builder = builders.RDS(self._config, self._resource, self._environment,
                               self._mappings, self._stack)
        self._template.update_resources(builder.resources)

    def _build_service_resources(self):
        service_path = path.join(self._config_path, self._resource_folder)
        builder = builders.Service(self._config, self._resource,
                                   self._environment, self._mappings,
                                   self._stack, self._region, self._amis,
                                   self._instances, service_path)
        self._template.update_resources(builder.resources)

    def _build_template_resources(self):
        self._template.update_mappings(self._mappings)
        if self._resource_type == 'environment':
            return self._build_environment_resources()

        self._template.update_mappings(self._stack.mappings)

        if self._resource_type == 'rds':
            self._build_rds_resources()

        elif self._resource_type == 'service':
            self._build_service_resources()

    def _error(self, message):
        """Write out an error message and exit.

        :param str message: The error message

        """
        sys.stderr.write('ERROR: {0}\n'.format(message))
        sys.exit(1)

    def _flatten_config(self, config):
        """Take a given config dictionary and if it contains environment
        specific values, map the environment values to the associated
        top level keys.

        :param dict config: The configuration to flatten
        :rtype: dict

        """
        output = {}
        for key, value in config.items():
            if isinstance(value, dict) and self._environment in value:
                output[key] = value[self._environment]
            else:
                output[key] = value
        return output

    def _get_stack(self):
        """Return a stack instance if the environment is not

        """
        if self._resource_type == 'environment':
            return
        return stack.Stack(self._environment, self._environment_config,
                           None, self._profile)

    def _load_config(self):
        """Return the config for the specified resource type

        :rtype: dict

        """
        if self._resource_type in ['environment', 'service']:
            config = self._load_config_file(self._resource_folder,
                                            self._resource_type)
        else:
            config = self._load_config_file(self._resource_folder,
                                            self._resource)
        return self._flatten_config(config)

    def _load_config_file(self, folder, file):
        """Return the contents of the specified configuration file in the
        specified configuration folder.


        :param str folder: The folder to load the configuration file from
        :param str file: The file to load
        :rtype: dict

        """
        file_path = path.join(self._config_path, folder,
                              '{}.yaml'.format(file))
        if path.exists(file_path):
            LOGGER.debug('Loading configuration from %s', file_path)
            with open(file_path, 'r') as handle:
                return yaml.load(handle)
        LOGGER.debug('Configuration file not found: %s', file_path)
        return {}

    def _load_environment_config(self):
        """Return the environment configuration

        :rtype: dict

        """
        if not self._environment:
            return {}
        return self._load_config_file(path.join(STACK_FOLDERS['environment'],
                                                self._environment),
                                      'environment')

    def _load_environment_mappings(self):
        """Return the mappings from the environment folder

        :rtype: dict

        """
        if not self._environment:
            return {}
        return self._load_config_file(path.join(STACK_FOLDERS['environment'],
                                                self._environment),
                                      'mappings')

    def _load_mappings(self):
        """Return the mapping data from the various config dirs and return
        merged mapping values in order of precedence of global, environment,
        or service/entity.

        :rtype: dict

        """
        mappings = dict()
        mappings.update(self._load_config_file(self._config_path, 'mappings'))
        if not self._resource == 'environment':
            mappings.update(self._load_environment_mappings())

        mappings.update(self._load_config_file(self._resource_folder,
                                               'mappings'))

        return mappings

    @staticmethod
    def _normalize_path(value): # pragma: no cover
        """Normalize the specified path value returning the absolute
        path for it.

        :param str value: The path value to normalize
        :rtype: str

        """
        return path.abspath(path.normpath(value))

    @property
    def _region(self):
        """Return the region value from the environment config,
        defaulting to us-east-1

        :rtype: str

        """
        return self._environment_config.get('region', 'us-east-1')

    @property
    def _resource_folder(self):
        """Return the folder that contains the resource's configuration data

        :rtype: str

        """
        if self._resource_type in ['environment', 'service']:
            return path.join('{}s'.format(self._resource_type),
                             self._resource)
        return self._resource_type

    @property
    def _template_name(self):
        """Return the template name based upon the

        :rtype: str

        """
        if self._environment:
            return '{0}-{1}-{2}'.format(self._environment, self._resource_type,
                                        self._resource)
        return self._resource

    @classmethod
    def _validate_arguments(cls, config_path, action, environment,
                            resource_type, resource):
        """Validate the initialization arguments, raising ``ValueError`` if any
        do not validate.

        :raises: ValueError

        """
        if not cls._validate_action(action):
            raise ValueError('Invalid action: {}'.format(action))

        if not cls._validate_resource_type(resource_type):
            raise ValueError('Invalid resource type: {}'.format(resource_type))

        if not cls._validate_config_path(config_path):
            raise ValueError('Invalid config path: {}'.format(config_path))

        if (resource_type != 'environment' and
                not cls._validate_environment(config_path, environment)):
            raise ValueError('Invalid environment: {}'.format(environment))

        if not cls._validate_resource(config_path, resource_type, resource):
            raise ValueError('Invalid resource: {}'.format(resource))

    @staticmethod
    def _validate_action(action):
        """Validate the action is a valid action

        :param str action: The action to validate
        :rtype: bool

        """
        return action in ACTIONS

    @staticmethod
    def _validate_config_path(config_path):
        """Validate that the specified configuration path exists and
        contains at least some of the files or folders expected.

        :param str config_path: The path to validate
        :rtype: bool

        """
        paths_found = [path.exists(config_path)]
        paths_found += [path.exists(path.join(config_path, f)) for f in
                        CONFIG_FILES]
        return any(paths_found)

    @staticmethod
    def _validate_environment(config_path, environment):
        """Validate that the expected environment ``network.yaml``
        configuration file exists within the config path for the environment.

        :param str config_path: The base config path
        :param str environment: The environment name
        :rtype: bool

        """
        return path.exists(path.join(config_path,
                                     STACK_FOLDERS['environment'],
                                     environment,
                                     'environment.yaml'))

    @staticmethod
    def _validate_resource(config_path, resource_type, resource):
        """Validate the resource configuration directory exists

        :param str resource_type: The resource type specified
        :param str resource: The resource to validate
        :rtype: bool

        """
        if resource_type in ['environment', 'service']:
            return path.exists(path.join(config_path,
                                         '{}s'.format(resource_type),
                                         resource,
                                         '{}.yaml'.format(resource_type)))
        return path.exists(path.join(config_path,
                                     STACK_FOLDERS[resource_type],
                                     '{}.yaml'.format(resource)))

    @staticmethod
    def _validate_resource_type(resource_type):
        """Validate the resource is valid

        :param str resource_type: The resource type specified
        :rtype: bool

        """
        return resource_type in RESOURCE_TYPES



"""
Cloud Formation ELB Resources

"""
from formulary import base


class ELB(base.Resource):
    def __init__(self, name, instances, health_check, listeners,
                 security_groups, subnets, policies=None, cross_zone=False,
                 internal=False):
        super(ELB, self).__init__('AWS::ElasticLoadBalancing::LoadBalancer')
        self._properties = {
            'CrossZone': cross_zone,
            'HealthCheck': health_check.as_dict(),
            'LoadBalancerName': name,
            'Instances': instances,
            'Listeners': [l.as_dict() for l in listeners],
            'Policies': policies,
            'Scheme': 'internal' if internal else 'internet-facing',
            'SecurityGroups': security_groups,
            'Subnets': subnets
        }


class ELBHeathCheck(base.Property):

    def __init__(self, interval=30, target='HTTP:80', timeout=5,
                 healthy=10, unhealthy=2):
        super(ELBHeathCheck, self).__init__()
        self._values = {
            'HealthyThreshold': healthy,
            'Interval': interval,
            'Target': target,
            'Timeout': timeout,
            'UnhealthyThreshold': unhealthy
        }


class ELBListener(base.Property):
    def __init__(self, port, protocol, instance_port, instance_protocol,
                 policy_names=None, ssl_certificate_id=None):
        super(ELBListener, self).__init__()
        self._values = {'InstancePort': instance_port,
                        'InstanceProtocol': instance_protocol,
                        'LoadBalancerPort': port,
                        'PolicyNames':  policy_names or [],
                        'Protocol': protocol,
                        'SSLCertificateId': ssl_certificate_id}





"""
Build the resources for a Security Group

"""
from formulary import base
from formulary import resources
from formulary import utils

class SecurityGroup(base.Builder):

    def __init__(self, config, name, environment, mappings, stack):
        super(SecurityGroup, self).__init__(config, name, environment, mappings)
        self._owner = name
        self._name = '{0}-{1}-security-group'.format(environment, name)
        self._stack = stack
        self.logical_id = self._add_security_group()

    def _add_security_group(self):
        if isinstance(self._config.get('security-group'), str):
            return self._config.get('security-group')

        desc = ('Security Group for the {0} '
                'service in {1}').format(self._owner.capitalize(),
                                         self._stack.environment.capitalize())
        resource = resources.SecurityGroup(self._name,
                                           desc,
                                           self._stack.vpc.id,
                                           self._build_ingress_rules())
        self._add_resource(self._name, resource)
        return self._name

    def _build_ingress_rules(self):
        rules = []
        group = self._config.get('security-group', {})
        ingress_rules = list(group.get('ingress', {}))
        for row in ingress_rules:
            try:
                port, source = dict(row).popitem()
            except KeyError:
                continue
            protocol, from_port, to_port = utils.parse_port_value(port)
            cidr_block = utils.find_in_map(source)
            rules.append(resources.SecurityGroupRule(protocol,
                                                     from_port, to_port,
                                                     cidr_block).as_dict())
        return rules



"""
Cloud Formation EC2 Resources

"""
from formulary import base


class EC2BlockDevice(base.Property):
    def __init__(self, name, ebs, no_device=None, virtual_name=None):
        super(EC2BlockDevice, self).__init__()
        self._values = {'DeviceName': name,
                        'Ebs': ebs,
                        'NoDevice': no_device or {},
                        'VirtualName': virtual_name}


class EC2Instance(base.Resource):
    def __init__(self, name, ami, availability_zone, instance_type, subnet,
                 security_group, user_data, storage_size=20, private_ip=None):
        super(EC2Instance, self).__init__('AWS::EC2::Instance')
        self._name = name
        self._subnet = subnet
        nic = {
            'AssociatePublicIpAddress': True,
            'DeviceIndex': '0',
            'GroupSet': [security_group],
            'SubnetId': subnet
        }
        if private_ip:
            nic['PrivateIpAddress'] = private_ip
        volume = {
            'DeviceName': '/dev/xvda',
            'Ebs': {'VolumeType': 'gp2',
                    'VolumeSize': storage_size}
        }
        self._properties = {
            'AvailabilityZone': availability_zone,
            'BlockDeviceMappings': [volume],
            'DisableApiTermination': False,
            'EbsOptimized': False,
            'ImageId': ami,
            'InstanceInitiatedShutdownBehavior': 'stop',
            'InstanceType': instance_type,
            'KeyName': {'Fn::FindInMap': ['AWS', 'KeyName', 'Value']},
            'Monitoring': False,
            'NetworkInterfaces': [nic],
            'UserData': user_data
        }

    @property
    def subnet(self):
        return self._subnet


class SecurityGroup(base.Resource):
    def __init__(self, name, description, vpc, ingress):
        super(SecurityGroup, self).__init__('AWS::EC2::SecurityGroup')
        self._name = name
        self._properties['GroupDescription'] = description
        self._properties['SecurityGroupIngress'] = ingress
        self._properties['VpcId'] = vpc



class SecurityGroupRule(base.Property):
    def __init__(self, protocol, from_port,
                 to_port=None,
                 cidr_addr=None,
                 source_id=None,
                 source_name=None,
                 source_owner=None):
        super(SecurityGroupRule, self).__init__()
        self._values = {'CidrIp': cidr_addr,
                        'FromPort': from_port,
                        'IpProtocol': protocol,
                        'SourceSecurityGroupId': source_id,
                        'SourceSecurityGroupName': source_name,
                        'SourceSecurityGroupOwnerId': source_owner,
                        'ToPort': to_port or from_port}




"""
Formulary: Cloud-Formation Stack Management

"""
__version__ = '0.1.1'

LOG_CONFIG = {
    'version': 1,
    'disable_existing_loggers': True,
    'incremental': False,
    'formatters': {
        'console': {
            'format': (
                '%(levelname)-8s %(name) -24s: %(message)s'
            )
        },
    },
    'handlers': {
        'console': {
            'class': 'logging.StreamHandler',
            'formatter': 'console',
        },
    },
    'loggers': {
        'boto': {
            'handlers': ['console'],
            'level': 'WARNING',
        },
        'formulary': {
            'handlers': ['console'],
            'level': 'DEBUG',
        }
    }
}




"""

"""
import base64
import math
from os import path
import random
import re

from formulary import base
from formulary import resources
from formulary.builders import securitygroup
from formulary import utils

DEFAULT_INSTANCE_TYPE = 't2.small'
USER_DATA_RE = re.compile(r'\{\^(?P<command>instance|map)\s(?P<key>[\w\.]+)\}')


class Service(base.Builder):

    def __init__(self, config, name, environment, mappings,
                 stack, region, amis, instances, local_path):
        super(Service, self).__init__(config, name, environment, mappings)
        self._amis = amis
        self._instances = instances
        self._local_path = local_path
        self._mappings = mappings
        self._mappings.update(stack.mappings)
        self._region = region
        self._stack = stack
        self._security_group = self._add_security_group()
        self._add_instances()
        self._maybe_add_elbs()
        self._add_tag_to_resources('Environment', self._stack.environment)
        self._add_tag_to_resources('Service', self._name)

    def _add_autobalanced_instances(self):
        count = self._config.get('instance-count', 1)
        subnets = self._get_subnets(count)
        for index in range(0, count):
            subnet = subnets.pop(0)
            self._add_instance('{0}{1}'.format(self._name, index),
                               subnet, self._config)

    def _add_elb(self, name, config):
        protocol = config.get('protocol', 'http')
        instance_port = config.get('instance_port', config['port'])
        kwargs = {
            'port': config['port'],
            'protocol': protocol,
            'instance_port': instance_port,
            'instance_protocol': config.get('instance_protocol', protocol)
        }
        listeners = [resources.ELBListener(**kwargs)]

        kwargs = {'interval': config.get('interval'),
                  'target': '{0}:{1}{2}'.format(protocol.upper(), instance_port,
                                                config.get('check', '')),
                  'timeout': config.get('timeout'),
                  'healthy': config.get('healthy'),
                  'unhealthy': config.get('unhealthy')}
        for key, value in [(k, v) for k, v in kwargs.items()]:
            if value is None:
                del kwargs[key]
        health_check = resources.ELBHeathCheck(**kwargs)

        instances = [{'Ref': i} for i in self._instance_ids]

        self._add_resource(name,
                           resources.ELB(name, instances,
                                         health_check, listeners,
                                         [{'Ref': self._security_group}],
                                         self._instance_subnets))

    def _add_instance(self, name, subnet, config):
        """Add an instance to the resources for the given name, subnet, and
        config.

        :param str name: The name of the instance
        :param formulary.records.Subnet subnet: The subnet config
        :param dict config: The instance config
        :rtype: str

        """
        kwargs = {
            'name': '{0}-service-{1}'.format(self._environment, name),
            'ami': self._get_ami_id(),
            'availability_zone': subnet.availability_zone,
            'instance_type': config.get('instance-type'),
            'environment': self._environment,
            'region': self._region,
            'subnet': subnet.id,
            'security_group': {'Ref': self._security_group},
            'service': self._name,
            'storage_size': config.get('storage-capacity'),
            'private_ip': config.get('private_ip', '')
        }
        kwargs['user_data'] = self._get_user_data(config.get('user-data'),
                                                  kwargs)
        for key in ['environment', 'region', 'service']:
            del kwargs[key]
        resource = resources.EC2Instance(**kwargs)
        return self._add_resource(name, resource)

    def _add_instances(self):
        if 'instances' in self._config:
            for name, instance_cfg in self._config['instances'].items():
                cfg = dict(self._config)
                del cfg['instances']
                cfg.update(instance_cfg)
                for key in ['availability_zone', 'private_ip']:
                    cfg[key] = self._maybe_replace_with_mapping(cfg[key])
                self._maybe_add_availability_zone(cfg)
                self._add_instance('{0}-{1}'.format(self._environment, name),
                                   self._get_subnet(cfg['availability_zone']),
                                   cfg)
        elif self._config.get('instance-strategy') == 'same-az':
            self._maybe_add_availability_zone(self._config)
            subnet = self._get_subnet(self._config['availability_zone'])
            for index in range(0, self._config.get('instance-count', 1)):
                self._add_instance('{0}-{1}{2}'.format(self._environment,
                                                       self._name, index),
                                   subnet, self._config)
        elif self._config.get('instance-strategy') == 'az-balanced':
            return self._add_autobalanced_instances()
        elif 'instance-strategy' in self._config:
            raise ValueError('Unknown instance-strategy: '
                             '{0}'.format(self._config['instance-strategy']))

    def _add_security_group(self):
        builder = securitygroup.SecurityGroup(self._config,
                                              self._name,
                                              self._environment,
                                              self._mappings,
                                              self._stack)
        self._resources.update(builder.resources)
        return utils.camel_case(builder.logical_id)

    def _add_tag_to_resources(self, tag, value):
        for name, resource in self._resources.items():
            resource.add_tag(tag, value)

    def _get_ami_id(self):
        try:
            return self._amis[self._region][self._config.get('ami')]
        except KeyError:
            raise ValueError('AMI "%s" not found' % self._config.get('ami'))

    def _get_subnet(self, availability_zone):
        for subnet in self._get_subnets(len(self._stack.subnets)):
            if subnet.availability_zone == availability_zone:
                return subnet

    def _get_subnets(self, count):
        subnets = self._stack.subnets
        if count > len(subnets):
            subnets *= math.ceil(count / float(len(subnets)))
        return subnets[0:count]

    def _get_user_data(self, filename, kwargs):
        if filename:
            with open(path.join(self._local_path, filename), 'r') as handle:
                return self._render_user_data(handle.read(), kwargs)
        return None

    @property
    def _instance_ids(self):
        ids = []
        for k, v in self._resources.items():
            if isinstance(v, resources.EC2Instance):
                ids.append(k)
        return ids

    @property
    def _instance_subnets(self):
        return [v.subnet for v in self._resources.values()
                if isinstance(v, resources.EC2Instance)]

    def _maybe_add_availability_zone(self, config):
        if 'availability_zone' not in config:
            offset = random.randint(0, len(self._stack.subnets) - 1)
            config['availability_zone'] = \
                self._stack.subnets[offset].availability_zone

    def _maybe_add_elbs(self):
        for name, config in self._config.get('elb', {}).items():
            self._add_elb('{0}-{1}'.format(self._environment, name), config)

    def _render_user_data(self, content, kwargs):
        mappings = dict(self._mappings)
        for match in USER_DATA_RE.finditer(content):
            if match.group(1) == 'map':
                value = mappings
                for key in str(match.group(2)).split('.'):
                    value = value.get(key.strip())
                content = content.replace(match.group(0), value)
            elif match.group(1) == 'instance':
                content = content.replace(match.group(0),
                                          kwargs[match.group(2)])
        return base64.b64encode(content.encode('utf-8')).decode('utf-8')



"""
Build the AWS VPC environment by adding the various resources to the
Cloud Formation template

"""

from formulary import base
from formulary import resources
from formulary import utils


class Environment(base.Builder):

    def __init__(self, config, name, mappings):
        super(Environment, self).__init__(config, name, None, mappings)
        self._vpc, self._vpc_name = self._add_vpc()
        self._add_dhcp()
        self._gateway = self._add_gateway()
        self._internet_gateway = self._add_gateway_attachment()
        self._route_table = self._add_route_table()
        self._add_public_route()
        self._acl = self._add_network_acl()
        self._add_network_acl_entries()
        self._add_subnets()

    def _add_vpc(self):
        """Add the VPC section to the template, returning the id and name
        for use in the other sections of the stack configuration

        :rtype: str, str

        """
        vpc_name = self._name.replace('_', '-')
        resource = resources.VPC(vpc_name,
                                 self._config['vpc']['dns-support'],
                                 self._config['vpc']['dns-hostnames'],
                                 self._config['CIDR'])
        resource.add_tag('Environment', self._config['environment'])
        return self._add_resource(vpc_name, resource), vpc_name

    def _add_dhcp(self):
        """Add all of the DHCP options to the template for the given VPC"""
        self._add_dhcp_association(self._add_dhcp_options())

    def _add_dhcp_options(self):
        """Add all of the DHCP options to the template for the given VPC

        :rtype: str

        """
        config = self._config['dhcp-options']
        options = '{0}-dhcp'.format(self._vpc_name)
        self._add_resource(options,
                           resources.DHCPOptions(config['domain-name'],
                                                 config['name-servers'],
                                                 config['ntp-servers']))
        return options

    def _add_dhcp_association(self, dhcp_id):
        """Add all of the DHCP OptionsAssociation to the template for the given
        VPC and DHCP Options ID.

        :param str dhcp_id: The DHCP Options ID

        """
        self._add_resource('{0}-dhcp-assoc'.format(dhcp_id),
                           resources.DHCPOptionsAssociation(
                               utils.camel_case(dhcp_id),
                               utils.camel_case(self._vpc_name)))

    def _add_gateway(self):
        """Add a gateway to the template for the specified VPC

        :rtype: str

        """
        gateway = '{0}-gateway'.format(self._vpc_name)
        self._add_resource(gateway, resources.Gateway())
        return gateway

    def _add_gateway_attachment(self):
        """Attach the specified gateway to the VPC

        :rtype: str

        """
        attachment = '{0}-attachment'.format(self._gateway)
        self._add_resource(attachment,
                           resources.GatewayAttachment(
                               utils.camel_case(self._vpc_name),
                               utils.camel_case(self._gateway)))
        return attachment

    def _add_network_acl(self):
        """Add the Network ACL to the VPC

        :rtype: str

        """
        resource = resources.NetworkACL(self._vpc_name, self._vpc)
        resource.add_tag('Environment', self._config['environment'])
        acl = '{0}-network-acl'.format(self._vpc_name)
        self._add_resource(acl, resource)
        return acl

    def _add_network_acl_entries(self):
        """Iterate through the ACL entries and add them"""
        for index, acl in enumerate(self._config['network-acls']):
            self._add_resource('{0}{1}'.format(self._acl, index),
                               resources.NetworkACLEntry(
                                   utils.camel_case(self._acl),
                                   acl['CIDR'], acl['number'], acl['action'],
                                   acl['egress'], acl['ports']))

    def _add_public_route(self):
        """Add the public route specified in the mapping ``pubic/cidr`` for
        the specified VPC, route table, gateway and internet gateway.

        """
        self._add_resource('{0}-route'.format(self._vpc_name),
                           resources.Route(
                               utils.camel_case(self._route_table),
                               {'Fn::FindInMap': ['SubnetConfig',
                                                  'Public',
                                                  'CIDR']},
                               utils.camel_case(self._gateway),
                               utils.camel_case(self._internet_gateway)))

    def _add_route_table(self,):
        """Add the the route table for the specified VPC

        :rtype: str

        """
        route_table = '{0}-route-table'.format(self._vpc_name)
        self._add_resource(route_table,
                           resources.RouteTable(
                               utils.camel_case(self._vpc_name)))
        return route_table

    def _add_subnets(self):
        """Add the network subnets for the specified VPC and route table"""
        subnet_ids = []
        for subnet in self._config['subnets']:
            config = self._config['subnets'][subnet]
            subnet_id = '{0}{1}-subnet'.format(self._vpc_name, subnet)
            subnet_ids.append(utils.camel_case(subnet_id))
            resource = resources.Subnet(self._vpc_name, subnet, self._vpc,
                                        config['availability_zone'],
                                        config['CIDR'])
            resource.add_tag('Environment', self._config['environment'])
            self._add_resource(subnet_id, resource)
            self._add_resource('{0}-assoc'.format(subnet_id),
                               resources.SubnetRouteTableAssociation(
                                   utils.camel_case(subnet_id),
                                   utils.camel_case((self._route_table))))



"""
Formulary Command Line Interface

"""
import argparse
import logging.config
from os import path
import sys

from formulary import controller
from formulary import LOG_CONFIG

DESCRIPTION = 'AWS Cloud Formation Stack Management'


class CLI(object):

    def __init__(self):
        self._parser = self._create_parser()

    def run(self):
        args = self._parser.parse_args()
        if args.verbose:
            logging.config.dictConfig(LOG_CONFIG)
        try:
            obj = controller.Controller(args.config_dir,
                                        args.command,
                                        args.environment,
                                        args.type,
                                        args.name,
                                        args.verbose,
                                        args.dry_run,
                                        args.profile)
        except ValueError as error:
            sys.stderr.write('{}\n'.format(error))
            sys.exit(1)

        obj.execute()

    @staticmethod
    def _add_commands(parser):
        commands = parser.add_subparsers(title='Commands', dest='command')
        create = commands.add_parser('create',
                                     help='Create a Cloud Formation stack')
        create.add_argument('type', choices=controller.RESOURCE_TYPES,
                            help='The type of stack to create')
        create.add_argument('name', help='The name of the stack to create')

        update = commands.add_parser('update',
                                     help='Update a Cloud Formation stack')
        update.add_argument('type', choices=controller.RESOURCE_TYPES,
                            help='The type of stack to update')
        update.add_argument('name', help='The name of the stack to update')

    def _create_parser(self):
        parser = argparse.ArgumentParser(description=DESCRIPTION)
        self._add_commands(parser)
        parser.add_argument('-c', '--config-dir',
                            default=path.abspath('.'),
                            help='Specify the path to the configuration '
                                 'directory. Default: .')
        parser.add_argument('-d', '--dry-run', action='store_true')
        parser.add_argument('-e', '--environment',
                            help='The formulary environment name')
        parser.add_argument('-p', '--profile',
                            help='The AWS credentials profile to use')
        parser.add_argument('-v', '--verbose', action='store_true')
        return parser


def run():
    CLI().run()



import setuptools

classifiers = ['Development Status :: 3 - Alpha',
               'Intended Audience :: System Administrators',
               'License :: OSI Approved :: BSD License',
               'Natural Language :: English',
               'Operating System :: OS Independent',
               'Programming Language :: Python :: 2',
               'Programming Language :: Python :: 2.6',
               'Programming Language :: Python :: 2.7',
               'Programming Language :: Python :: 3',
               'Programming Language :: Python :: 3.2',
               'Programming Language :: Python :: 3.3',
               'Programming Language :: Python :: 3.4',
               'Programming Language :: Python :: Implementation :: CPython',
               'Programming Language :: Python :: Implementation :: PyPy',
               'Topic :: System :: Clustering',
               'Topic :: System :: Installation/Setup',
               'Topic :: System :: Networking',
               'Topic :: System :: Systems Administration']

setuptools.setup(name='formulary',
                 version='0.1.1',
                 description='Easy management of AWS Cloud Formation stacks',
                 long_description=open('README.rst').read(),
                 author='Gavin M. Roy',
                 author_email='gavinr@aweber.com',
                 url='http://formulary.readthedocs.org',
                 packages=['formulary', 'formulary.builders', 'formulary.resources'],
                 package_data={'': ['LICENSE',
                                    'README.rst',
                                    'requirements.txt']},
                 include_package_data=True,
                 install_requires=open('requirements.txt', 'r').read(),
                 license='BSD',
                 classifiers=classifiers,
                 entry_points={'console_scripts':
                                   ['formulary=formulary.cli:run']},
                 zip_safe=True)



"""
namedtuple representations of Stack resources

"""
from collections import namedtuple

DHCPOptions = namedtuple('DHCPOptions', ['id', 'logical_id', 'timestamp',
                                         'status', 'domain_name',
                                         'dns_servers', 'ntp_servers', 'tags'])

DHCPOptionsAssociation = namedtuple('DHCPOptionsAssociation',
                                    ['id', 'logical_id', 'timestamp',
                                     'status'])

GatewayAssociation = namedtuple('GatewayAssociation',
                                ['id', 'logical_id', 'timestamp', 'status'])


InternetGateway = namedtuple('InternetGateway',
                             ['id', 'logical_id', 'timestamp', 'status'])

NetworkACL = namedtuple('NetworkACL', ['id', 'logical_id', 'timestamp',
                                       'status', 'entries', 'tags'])

NetworkACLEntry = namedtuple('NetworkACLEntry', ['logical_id', 'timestamp',
                                                 'status', 'number',
                                                 'cidr_block', 'egress',
                                                 'action', 'protocol',
                                                 'from_port', 'to_port'])

Route = namedtuple('Route', ['id', 'logical_id', 'timestamp', 'status'])

RouteTable = namedtuple('RouteTable', ['id', 'logical_id', 'timestamp',
                                       'status', 'associations', 'tags'])

Subnet = namedtuple('Subnet', ['id',  'logical_id', 'timestamp', 'status',
                               'availability_zone', 'cidr_block',
                               'available_ips'])

SubnetRouteTableAssociation = namedtuple('SubnetRouteTableAssociation',
                                         ['id', 'logical_id', 'timestamp',
                                          'status'])

VPC = namedtuple('VPC', ['id', 'logical_id', 'timestamp', 'status',
                         'description', 'cidr_block', 'is_default',
                         'instance_tenancy', 'dhcp_options_id', 'tags'])



"""
Cloud FormationStack

"""
import json
import logging

from boto import cloudformation
from boto import vpc

from formulary import records
from formulary import utils

LOGGER = logging.getLogger(__name__)

class _API(object):

    def __init__(self, region, profile):
        kwargs = {'profile_name': profile} if profile else {}
        LOGGER.debug('Connecting with profile: %r', kwargs)
        self._cf = cloudformation.connect_to_region(region, **kwargs)
        self._vpc = vpc.connect_to_region(region, **kwargs)

    def close(self):
        self._cf.close()
        self._vpc.close()

    @property
    def cf(self):
        """Return the connection handle to the Cloud Formation API

        :rtype: boto.cloudformation.connection.CloudFormationConnection

        """
        return self._cf

    @property
    def vpc(self):
        """Return the connection handle to the VPC API

        :rtype: boto.vpc.VPCConnection

        """
        return self._vpc

class Stack(object):

    def __init__(self, name, config, template=None, profile=None):
        self._config = config
        self._name = name
        self._profile = profile
        self._region = config.get('region')

        self._api = None
        self._description = None
        self._id = None
        self._resources = {}
        self._status = None
        if template:
            self._process_template(template)
        else:
            self._process_api_data()

    @property
    def cidr_block(self):
        """Return the CIDR block for the Stack's VPC

        :rtype: str

        """
        return self.vpc.cidr_block

    @property
    def environment(self):
        """Return the environment from the Stack's tags

        :rtype: str

        """
        return self.vpc.tags.get('Environment', self._name)

    @property
    def mappings(self):
        """Return mappings based upon values in the stack used for service
        configuration and user-data templates.

        :rtype: dict

        """
        vpc_data = dict()
        for key, value in self.vpc._asdict().items():
            if key == 'tags':
                continue
            cckey = utils.camel_case(key)
            if key == 'cidr_block':
                cckey = 'CIDR'
            if value is not None:
                vpc_data[cckey] = value
        return {
            'Network': {
                'Name': {'Value': self._name},
                'VPC': vpc_data,
                'AWS': {'Region': self._region}
            }
        }

    @property
    def resources(self):
        """Return the resources associated with this stack

        :rtype: list

        """
        return self._resources

    @property
    def subnets(self):
        """Return a list of subnets associated with the stack

        :rtype: list of formulary.records.Subnet

        """
        return [r for r in self._resources if isinstance(r, records.Subnet)]

    @property
    def vpc(self):
        """Return the VPC if it is in the resources.

        :rtype: formulary.records.VPC

        """
        for resource in self._resources:
            if isinstance(resource, records.VPC):
                return resource

    def _build_api_create_function_map(self):
        return {
            'AWS::EC2::DHCPOptions': self._create_dhcp_options_tuple,
            'AWS::EC2::VPCDHCPOptionsAssociation':
                self._create_dhcp_options_assoc_tuple,
            'AWS::EC2::InternetGateway': self._create_internet_gateway_tuple,
            'AWS::EC2::NetworkAcl': self._create_network_acl_tuple,
            'AWS::EC2::NetworkAclEntry': self._create_network_acl_tuple_entry,
            'AWS::EC2::Route': self._create_route_tuple,
            'AWS::EC2::RouteTable': self._create_route_table_tuple,
            'AWS::EC2::Subnet': self._create_subnet_tuple,
            'AWS::EC2::SubnetRouteTableAssociation':
                self._create_subnet_route_table_association,
            'AWS::EC2::VPC': self._create_vpc_tuple,
            'AWS::EC2::VPCGatewayAttachment':
                self._create_gateway_association_tuple
        }

    def _create_dhcp_options_tuple(self, value):
        """Create the DHCPOptions tuple from the retrieved stack resource
        information

        :param value: The DHCPOptions stack entry
        :type value: boto.cloudformation.stack.StackResource
        :rtype: formulary.records.DHCPOptions

        """
        record = self._describe_dhcp_options(value.physical_resource_id)
        options = record.options
        domain = options['domain-name'][0] if options['domain-name'] else None
        return records.DHCPOptions(value.physical_resource_id,
                                   value.logical_resource_id,
                                   value.timestamp.isoformat(),
                                   value.resource_status,
                                   domain,
                                   options['domain-name-servers'][0],
                                   options['ntp-servers'][0],
                                   record.tags)

    @staticmethod
    def _create_dhcp_options_assoc_tuple(value):
        """Create the DHCPOptionsAssociation tuple from the retrieved stack
        resource information

        :param value: The DHCPOptionsAssociation stack entry
        :type value: boto.cloudformation.stack.StackResource
        :rtype: formulary.records.DHCPOptionsAssociation

        """
        return records.DHCPOptionsAssociation(value.physical_resource_id,
                                              value.logical_resource_id,
                                              value.timestamp.isoformat(),
                                              value.resource_status)

    @staticmethod
    def _create_gateway_association_tuple(value):
        """Create the GatewayAssociation tuple from the retrieved stack
        resource information

        :param value: The gateway association stack entry
        :type value: boto.cloudformation.stack.StackResource
        :rtype: formulary.records.GatewayAssociation

        """
        return records.GatewayAssociation(value.physical_resource_id,
                                          value.logical_resource_id,
                                          value.timestamp.isoformat(),
                                          value.resource_status)

    @staticmethod
    def _create_internet_gateway_tuple(value):
        """Create the InternetGateway tuple from the retrieved stack resource
        information

        :param value: The internet gateway stack entry
        :type value: boto.cloudformation.stack.StackResource
        :rtype: formulary.records.InternetGateway

        """
        return records.InternetGateway(value.physical_resource_id,
                                       value.logical_resource_id,
                                       value.timestamp.isoformat(),
                                       value.resource_status)

    def _create_network_acl_tuple(self, value):
        """Create the NetworkACL tuple from the retrieved stack resource
        information

        :param value: The NetworkACL stack entry
        :type value: boto.cloudformation.stack.StackResource
        :rtype: formulary.records.NetworkACL

        """
        acl_data = self._describe_network_acl(value.physical_resource_id)
        entries = []
        for entry in acl_data.network_acl_entries:
            from_port = int(entry.port_range.from_port) \
                if entry.port_range.from_port else None
            to_port = int(entry.port_range.to_port) \
                if entry.port_range.to_port else None
            acl_entry = records.NetworkACLEntry(None, None, None,
                                                int(entry.rule_number),
                                                entry.cidr_block,
                                                bool(entry.egress),
                                                entry.rule_action,
                                                int(entry.protocol),
                                                from_port, to_port)
            entries.append(acl_entry)
        return records.NetworkACL(value.physical_resource_id,
                                  value.logical_resource_id,
                                  value.timestamp.isoformat(),
                                  value.resource_status,
                                  entries,
                                  acl_data.tags)

    @staticmethod
    def _create_network_acl_tuple_entry(value):
        """Create the NetworkACLEntry tuple from the retrieved stack resource
        information

        :param value: The Network ACL entry stack entry
        :type value: boto.cloudformation.stack.StackResource
        :rtype: formulary.records.NetworkACL

        """
        return records.NetworkACLEntry(value.logical_resource_id,
                                       value.timestamp.isoformat(),
                                       value.resource_status,
                                       None, None, None, None, None, None, None)

    @staticmethod
    def _create_route_tuple(value):
        """Create the Route tuple from the retrieved stack resource
        information

        :param value: The route stack entry
        :type value: boto.cloudformation.stack.StackResource
        :rtype: formulary.records.Route

        """
        return records.Route(value.physical_resource_id,
                             value.logical_resource_id,
                             value.timestamp.isoformat(),
                             value.resource_status)

    def _create_route_table_tuple(self, value):
        """Create the RouteTable tuple from the retrieved stack resource
        information

        :param value: The route table stack entry
        :type value: boto.cloudformation.stack.StackResource
        :rtype: formulary.records.RouteTable

        """
        data = self._describe_route_table(value.physical_resource_id)
        associations = [d.subnet_id for d in data.associations]
        return records.RouteTable(value.physical_resource_id,
                                  value.logical_resource_id,
                                  value.timestamp.isoformat(),
                                  value.resource_status,
                                  associations,
                                  data.tags)

    def _create_subnet_tuple(self, value):
        """Create the Subnet tuple from the retrieved stack resource
        information

        :param value: The subnet stack entry
        :type value: boto.cloudformation.stack.StackResource
        :rtype: formulary.records.Subnet

        """
        data = self._describe_subnet(value.physical_resource_id)
        return records.Subnet(value.physical_resource_id,
                              value.logical_resource_id,
                              value.timestamp.isoformat(),
                              value.resource_status,
                              data.availability_zone,
                              data.cidr_block,
                              data.available_ip_address_count)

    @staticmethod
    def _create_subnet_route_table_association(value):
        """Create the SubnetRouteTableAssociation tuple from the retrieved
        stack resource information

        :param value: The route stack entry
        :type value: boto.cloudformation.stack.StackResource
        :rtype: formulary.records.SubnetRouteTableAssociation

        """
        return records.SubnetRouteTableAssociation(value.physical_resource_id,
                                                   value.logical_resource_id,
                                                   value.timestamp.isoformat(),
                                                   value.resource_status)

    def _create_vpc_tuple(self, value):
        """Create the VPC tuple from the retrieved stack resource information

        :param value: The VPC stack entry
        :type value: boto.cloudformation.stack.StackResource
        :return: formulary.records.VPC

        """
        vpc_data = self._describe_vpc(value.physical_resource_id)
        return records.VPC(value.physical_resource_id,
                           value.logical_resource_id,
                           value.timestamp.isoformat(),
                           vpc_data.state if vpc_data else None,
                           value.description,
                           vpc_data.cidr_block if vpc_data else None,
                           vpc_data.is_default if vpc_data else None,
                           vpc_data.instance_tenancy if vpc_data else None,
                           vpc_data.dhcp_options_id if vpc_data else None,
                           vpc_data.tags if vpc_data else None)

    def _describe_dhcp_options(self, dhcp_id):
        """Get the data from AWS about the specified DHCP options

        :param str dhcp_id: The ID to lookup
        :rtype: boto.vpc.dhcpoptions.DhcpOptions

        """
        LOGGER.debug('Fetching data for %s', dhcp_id)
        results = self._api.vpc.get_all_dhcp_options(dhcp_id)
        return results[0] if results else None

    def _describe_internet_gateway(self, gateway_id):
        """Get the data from AWS about the specified internet gateway

        :param str gateway_id: The ID to lookup
        :rtype: boto.vpc.internetgateway.InternetGateway

        """
        LOGGER.debug('Fetching data for %s', gateway_id)
        results = self._api.vpc.get_all_internet_gateways(gateway_id)
        return results[0] if results else None

    def _describe_route_table(self, route_table_id):
        """Get the data from AWS about the specified route table

        :param str route_table_id: The ID to lookup
        :rtype: boto.vpc.routetable.RouteTable

        """
        LOGGER.debug('Fetching data for %s', route_table_id)
        results = self._api.vpc.get_all_route_tables(route_table_id)
        return results[0] if results else None

    def _describe_stack(self):
        """Fetch the core stack configuration data

        :rtype: boto.cloudformation.stack.Stack

        """
        LOGGER.debug('Fetching data for %s', self._name)
        stacks = self._api.cf.describe_stacks(self._name)
        if not stacks:
            raise ValueError('Environment "{0}" not found'.format(self._name))
        return stacks[0]

    def _describe_stack_resources(self):
        """Fetch the resources for the stack

        :rtype: list

        """
        function_map = self._build_api_create_function_map()

        LOGGER.debug('Fetching stack resources')
        resources = []
        result = self._api.cf.describe_stack_resources(self._name)
        for row in result:
            try:
                resources.append(function_map[row.resource_type](row))
            except KeyError:
                LOGGER.debug('Row data: %r', row.__dict__)
                raise NotImplementedError(row.resource_type)
        return resources

    def _describe_network_acl(self, acl_id):
        """Get the data from AWS about the network ACL

        :param str acl_id: The ID to lookup
        :rtype: boto.vpc.networkacl.NetworkAcl

        """
        LOGGER.debug('Fetching data for %s', acl_id)
        results = self._api.vpc.get_all_network_acls(acl_id)
        return results[0] if results else None

    def _describe_subnet(self, subnet_id):
        """Fetch the subnet configuration data

        :rtype: boto.vpc.subnet.Subnet

        """
        LOGGER.debug('Fetching data for %s', subnet_id)
        results = self._api.vpc.get_all_subnets(subnet_id)
        return results[0] if results else None

    def _describe_vpc(self, vpc_id):
        """Get the data from AWS about the VPC

        :param str vpc_id: The ID to lookup
        :rtype: boto.vpc.vpc.VPC

        """
        LOGGER.debug('Fetching data for %s', vpc_id)
        results = self._api.vpc.get_all_vpcs(vpc_id)
        return results[0] if results else None

    def _process_api_data(self):
        """Build the stack resources from a combination of remote API calls
        to various AWS endpoints.

        """
        LOGGER.debug('Processing API data')
        self._api = _API(self._region, self._profile)
        stack = self._describe_stack()
        self._description = stack.description
        self._id = stack.stack_id
        self._status = stack.stack_status
        self._resources = self._describe_stack_resources()
        self._api.close()

    def _process_template(self, template):
        values = json.loads(template.as_json())
        self._description = values['Description']
        # @TODO add resource tuples



"""
Build the resources for a RDS Instance

"""
from formulary import base
from formulary import resources
from formulary.builders import securitygroup
from formulary import utils

DEFAULT_BACKUP_RETENTION = 3
DEFAULT_INSTANCE_TYPE = 'db.t2.small'
DEFAULT_ENGINE_TYPE = 'postgres'
DEFAULT_ENGINE_VERSION = '9.4.1'
DEFAULT_ENGINE_PORT = 5432
DEFAULT_STORAGE_CAPACITY = 100
DEFAULT_IOPS = 1000


class RDS(base.Builder):

    def __init__(self, config, name, environment, mappings, stack):
        super(RDS, self).__init__(config, name, environment, mappings)
        self._stack = stack
        self._security_group = self._add_security_group()
        self._add_instance()

    def _add_instance(self):
        subnets = self._add_subnet_groups(self._config)
        if not self._config.get('multi-az'):
            self._config['availability_zone'] = \
                self._stack.subnets[0].availability_zone
        name = '{0}-{1}'.format(self._environment, self._name)
        resource = resources.DBInstance(name, self._config,
                                        utils.camel_case(subnets),
                                        {'Ref': self._security_group})
        self._add_resource(name, resource)

    def _add_security_group(self):
        builder = securitygroup.SecurityGroup(self._config,
                                              self._name,
                                              self._environment,
                                              self._mappings,
                                              self._stack)
        self._resources.update(builder.resources)
        return utils.camel_case(builder.logical_id)

    def _add_subnet_groups(self, config):
        subnets = []
        if config.get('multi-az'):
            subnets.append(self._stack.subnets[0].id)
        else:
            for subnet in self._stack.subnets:
                subnets.append(subnet.id)
        subnet_group_name = '{0}-{1}-subnet-group'.format(self._environment,
                                                          self._name)
        resource = resources.DBSubnetGroup(subnet_group_name, subnets)
        self._add_resource(subnet_group_name, resource)
        return subnet_group_name



"""
Cloud Formation Resources

"""
from formulary.resources.ec2 import EC2Instance
from formulary.resources.ec2 import SecurityGroup
from formulary.resources.ec2 import SecurityGroupRule
from formulary.resources.elb import ELB
from formulary.resources.elb import ELBHeathCheck
from formulary.resources.elb import ELBListener
from formulary.resources.rds import DBInstance
from formulary.resources.rds import DBSubnetGroup
from formulary.resources.vpc import DHCPOptions
from formulary.resources.vpc import DHCPOptionsAssociation
from formulary.resources.vpc import Gateway
from formulary.resources.vpc import GatewayAttachment
from formulary.resources.vpc import NetworkACL
from formulary.resources.vpc import NetworkACLEntry
from formulary.resources.vpc import Route
from formulary.resources.vpc import RouteTable
from formulary.resources.vpc import Subnet
from formulary.resources.vpc import SubnetRouteTableAssociation
from formulary.resources.vpc import VPC



"""
Cloud Formation methods and interfaces

"""
import collections
import json
import logging

from boto import cloudformation
from boto import exception

LOGGER = logging.getLogger(__name__)

StackResource = collections.namedtuple('StackResource',
                                       ('id', 'type', 'name', 'status'))


def create_stack(region, template, profile):
    """Create a stack in the specified region with the given template.

    :param str region: The region name to create the stack in
    :param Template template: The template to use
    :param str profile: The credentials profile to use
    :raises: RequestException

    """
    kwargs = {'profile_name': profile} if profile else {}
    connection = cloudformation.connect_to_region(region, **kwargs)
    template_body = template.as_json()
    if not connection.validate_template(template_body):
        raise RequestException('The specified template did not validate')
    try:
        connection.create_stack(template.name, template_body)
    except exception.BotoServerError as error:
        raise RequestException(error)
    connection.close()


def estimate_stack_cost(region, template, profile):
    """Estimate the cost of the stack in EC2

    :param str region: The region name to create the stack in
    :param Template template: The template to use
    :param str profile: The credentials profile to use
    :raises: RequestException

    """
    kwargs = {'profile_name': profile} if profile else {}
    connection = cloudformation.connect_to_region(region, **kwargs)
    template_body = template.as_json()
    if not connection.validate_template(template_body):
        raise RequestException('The specified template did not validate')
    try:
        result = connection.estimate_template_cost(template_body)
    except exception.BotoServerError as error:
        raise RequestException(error)
    connection.close()
    return result


def update_stack(region, template, profile):
    """Update a stack in the specified region with the given template.

    :param str region: The region name to create the stack in
    :param Template template: The template to use
    :param str profile: The credentials profile to use
    :raises: RequestException

    """
    kwargs = {'profile_name': profile} if profile else {}
    connection = cloudformation.connect_to_region(region, **kwargs)
    template_body = template.as_json()
    if not connection.validate_template(template_body):
        raise RequestException('The specified template did not validate')
    try:
        connection.update_stack(template.name, template_body)
    except exception.BotoServerError as error:
        raise RequestException(error)
    connection.close()


class RequestException(Exception):
    def __init__(self, error):
        error = str(error)
        payload = json.loads(error[error.find('{'):])
        self._message = payload['Error']['Message']

    def __repr__(self):
        return '<{0} "{1}">'.format(self._message)

    def __str__(self):
        return self._message
